package com.softslate.commerce.businessobjects.shipping;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TimeZone;

import org.apache.axis.types.NonNegativeInteger;
import org.apache.axis.types.PositiveInteger;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.fedex.rate.stub.Address;
import com.fedex.rate.stub.ClientDetail;
import com.fedex.rate.stub.Contact;
import com.fedex.rate.stub.Dimensions;
import com.fedex.rate.stub.DropoffType;
import com.fedex.rate.stub.LinearUnits;
import com.fedex.rate.stub.Notification;
import com.fedex.rate.stub.NotificationSeverityType;
import com.fedex.rate.stub.PackageSpecialServiceType;
import com.fedex.rate.stub.PackageSpecialServicesRequested;
import com.fedex.rate.stub.PackagingType;
import com.fedex.rate.stub.Party;
import com.fedex.rate.stub.Payment;
import com.fedex.rate.stub.PaymentType;
import com.fedex.rate.stub.RatePortType;
import com.fedex.rate.stub.RateReply;
import com.fedex.rate.stub.RateReplyDetail;
import com.fedex.rate.stub.RateRequest;
import com.fedex.rate.stub.RateRequestType;
import com.fedex.rate.stub.RateServiceLocator;
import com.fedex.rate.stub.RatedShipmentDetail;
import com.fedex.rate.stub.RequestedPackageLineItem;
import com.fedex.rate.stub.RequestedShipment;
import com.fedex.rate.stub.ShipmentRateDetail;
import com.fedex.rate.stub.SignatureOptionDetail;
import com.fedex.rate.stub.SignatureOptionType;
import com.fedex.rate.stub.SmartPostIndiciaType;
import com.fedex.rate.stub.SmartPostShipmentDetail;
import com.fedex.rate.stub.TransactionDetail;
import com.fedex.rate.stub.TransitTimeType;
import com.fedex.rate.stub.VersionId;
import com.fedex.rate.stub.WebAuthenticationCredential;
import com.fedex.rate.stub.WebAuthenticationDetail;
import com.fedex.rate.stub.Weight;
import com.fedex.rate.stub.WeightUnits;
import com.softslate.commerce.businessobjects.core.ApplicationException;
import com.softslate.commerce.businessobjects.order.OrderShippingRule;
import com.softslate.commerce.businessobjects.product.Manufacturer;

/**
 * Connects to the FedEx server for live online rates. Sends Web Service rate
 * requests to the FedEx server and reads the responses from it. From the
 * response, it builds a set of shipping options which get presented to the
 * customer during checkout.
 * 
 * @author David Tobey
 * @see com.softslate.commerce.businessobjects.shipping.ShippingProcessor
 * @see com.softslate.commerce.businessobjects.shipping.BaseShippingProcessor
 * 
 */
public class FedExShippingProcessor extends BaseShippingProcessor implements
		ShippingProcessor {

	static Log log = LogFactory.getLog(FedExShippingProcessor.class);

	public Collection loadShippingOptions() throws Exception {
		if (log.isInfoEnabled())
			log.info("Loading shipping options.");

		// Throw an error if a delivery address is not defined.
		if (getUser().getOrder() == null
				|| getUser().getOrder().firstDelivery() == null
				|| getUser().getOrder().firstDelivery().getCountry() == null
				|| getUser().getOrder().firstDelivery().getState() == null
				|| getUser().getOrder().firstDelivery().getPostalCode() == null) {
			if (log.isInfoEnabled())
				log.info("Delivery address not defined.");
			String errorCode = "ADDRESS";
			String errorMessage = "Delivery address not defined.";
			throw new ApplicationException(errorCode + ": " + errorMessage);
		}

		// Throw an error if a delivery address is not defined.
		String missingRequiredSettings = missingRequiredSettings();
		if (missingRequiredSettings != null
				&& missingRequiredSettings.length() > 0) {
			if (log.isInfoEnabled())
				log.info("Not all required settings available: "
						+ missingRequiredSettings);
			String errorCode = "SETTINGS";
			String errorMessage = "Not all required settings available. Administrators: check required settings: "
					+ missingRequiredSettings;
			throw new ApplicationException(errorCode + ": " + errorMessage);
		}

		// A collection where all the rates from each drop shipper will be added
		// up
		Collection mergedRates = new ArrayList();

		// Find all the drop shippers represented by items in the order, and
		// define the dropShipperIDs variable.
		retrieveDropShippersFromOrder(getUser().getOrder());

		// First query for the shipping rates for everything that does not have
		// a drop shipper, using the global settings.
		if (log.isInfoEnabled())
			log.info("Processing non-drop shipped items.");
		Map parameters = new HashMap();
		parameters.put(
				"fedExOriginAddressStateOrProvinceCode",
				(String) getSettings().getValue(
						"fedExOriginAddressStateOrProvinceCode"));
		parameters.put("fedExOriginAddressPostalCode", (String) getSettings()
				.getValue("fedExOriginAddressPostalCode"));
		parameters.put("fedExOriginAddressCountryCode", (String) getSettings()
				.getValue("fedExOriginAddressCountryCode"));
		currentDropShipperID = null;
		Map ratesAndErrors = requestRates(parameters);

		// Skip if there were no items with the default drop shipper. Avoids "Invalid piece count" reply from FedEx.
		if (ratesAndErrors != null) {
			Collection rates = (Collection) ratesAndErrors.get("rates");
			String errorCode = (String) ratesAndErrors.get("errorCode");
			String errorMessage = (String) ratesAndErrors.get("errorMessage");
			if ((rates == null || rates.isEmpty()) && errorCode != null && errorMessage != null) {
				throw new ApplicationException(errorCode + ": " + errorMessage);
			}

			// Merge the rates for all drop shippers by adding them up
			mergedRates = mergeRates(mergedRates, rates);
		}

		// Query for each drop shipper now separately, adding the rates together
		// as we go.
		Iterator i = dropShippers.iterator();
		while (i.hasNext()) {
			Manufacturer m = (Manufacturer) i.next();
			parameters.put("fedExOriginAddressStateOrProvinceCode",
					m.getState());
			parameters.put("fedExOriginAddressPostalCode", m.getPostalCode());
			parameters.put("fedExOriginAddressCountryCode", m.getCountry());
			currentDropShipperID = new Integer(m.getManufacturerID());
			if (log.isInfoEnabled())
				log.info("Processing drop shipper: " + m.getCode());
			Map dropShipperRatesAndErrors = ratesAndErrors = requestRates(parameters);

			// Throw an exception if there was an error and no rates
			// returned
			Collection dropShipperRates = (Collection) dropShipperRatesAndErrors
					.get("rates");
			String dropShipperErrorCode = (String) dropShipperRatesAndErrors
					.get("errorCode");
			String dropShipperErrorMessage = (String) dropShipperRatesAndErrors
					.get("errorMessage");
			if ((dropShipperRates == null || dropShipperRates.isEmpty())
					&& dropShipperErrorCode != null
					&& dropShipperErrorMessage != null) {
				throw new ApplicationException(dropShipperErrorCode + ": "
						+ dropShipperErrorMessage);
			}

			// Merge the rates for all drop shippers by adding them up
			mergedRates = mergeRates(mergedRates, dropShipperRates);
		}

		// Apply the extra charge as defined in the settings
		mergedRates = doGlobalExtraCharge(mergedRates);
		return mergedRates;
	}

	public int processShipping() throws Exception {
		return 0;
	}

	public Map requestRates(Map parameters) throws Exception {
		Map ratesAndErrors = new HashMap();

		// Initialize the service
		RateServiceLocator service = new RateServiceLocator();
		RatePortType port;
		updateEndPoint(service);
		port = service.getRateServicePort();

		// Build the request
		RateRequest request = buildRateRequest(parameters);
		if (request == null) {
			return null;
		}

		// Send the request
		RateReply reply = port.getRates(request);

		// Process the response
		if (isResponseOk(reply.getHighestSeverity()) || isResponseBenignError(reply)) {
			ratesAndErrors = processRateResponse(parameters, reply);
		} else {

			// Process errors
			ratesAndErrors = processErrors(reply.getNotifications());
		}
		return ratesAndErrors;
	}

	private RateRequest buildRateRequest(Map parameters) {
		RateRequest request = new RateRequest();
		request.setClientDetail(createClientDetail());
		request.setWebAuthenticationDetail(createWebAuthenticationDetail());
		TransactionDetail transactionDetail = new TransactionDetail();
		transactionDetail.setCustomerTransactionId(getUser().getOrder()
				.getOrderID()
				+ "-"
				+ java.util.Calendar.getInstance().getTimeInMillis());
		request.setTransactionDetail(transactionDetail);
		VersionId versionId = new VersionId("crs", 14, 0, 0);
		request.setVersion(versionId);
		RequestedShipment requestedShipment = buildRequestedShipment(parameters);
		if (requestedShipment == null) {
			return null;
		}
		request.setRequestedShipment(requestedShipment);
		request.setReturnTransitAndCommit(Boolean.TRUE);
		return request;
	}

	public RequestedShipment buildRequestedShipment(Map parameters) {
		RequestedShipment requestedShipment = new RequestedShipment();
		requestedShipment.setShipTimestamp(createTimestampWithOffsetWorkaround());
		requestedShipment.setDropoffType(createDropoffType());
		requestedShipment.setPackagingType(createPackagingType());
		requestedShipment.setShipper(createShipperParty(parameters));
		requestedShipment.setRecipient(createRecipientParty(parameters));
		requestedShipment.setShippingChargesPayment(createShippingChargesPayment());
		PackageSpecialServicesRequested pssr = createSignatureOption();
		Collection packages = createPackages(parameters);

		// If no packages were found, return null to skip the call to FedEx
		if (packages == null || packages.isEmpty()) {
			return null;
		}

		// Put the packages all together
		String fedExShowListPrices = (String) getSettings().getValue(
				"fedExShowListPrices");
		int numberOfPackages = packages.size();
		RequestedPackageLineItem[] requestedPackages = new RequestedPackageLineItem[numberOfPackages];
		RateRequestType rateRequestType[] = new RateRequestType[numberOfPackages];
		double totalWeight = 0;
		WeightUnits totalWeightUnits = WeightUnits.LB;
		int count = 0;
		Iterator i = packages.iterator();
		while (i.hasNext()) {
			RequestedPackageLineItem pack = (RequestedPackageLineItem) i.next();
			pack.setSequenceNumber(new PositiveInteger(Integer
					.toString(count + 1)));
			pack.setSpecialServicesRequested(pssr);
			pack.setGroupPackageCount(new NonNegativeInteger("1"));
			requestedPackages[count] = pack;
			totalWeight = totalWeight
					+ pack.getWeight().getValue().doubleValue();
			totalWeightUnits = pack.getWeight().getUnits();
			rateRequestType[count] = RateRequestType.LIST;
			if (fedExShowListPrices != null && fedExShowListPrices.equals("0")) {
				rateRequestType[count] = RateRequestType.ACCOUNT;
			}
			count++;
		}
		requestedShipment.setRateRequestTypes(rateRequestType);
		requestedShipment.setRequestedPackageLineItems(requestedPackages);
		requestedShipment.setPackageCount(new NonNegativeInteger(Integer
				.toString(numberOfPackages)));

		// Total weight information
		Weight weight = new Weight();
		weight.setValue(new BigDecimal(totalWeight));
		weight.setUnits(totalWeightUnits);
		requestedShipment.setTotalWeight(weight);

		// Smart Post
		String fedexSmartPostIndiciaType = getSettings().getValue("fedexSmartPostIndiciaType");
		String fedexSmartPostHubId = getSettings().getValue("fedexSmartPostHubId");
		if (!StringUtils.isEmpty(fedexSmartPostIndiciaType) && !StringUtils.isEmpty(fedexSmartPostHubId)) {
			requestedShipment.setSmartPostDetail(createSmartPostShipmentDetail(fedexSmartPostIndiciaType,
					fedexSmartPostHubId));
		}

		return requestedShipment;
	}

	protected Calendar createTimestampWithOffsetWorkaround() {
		return createTimestamp(true);
	}

	protected Calendar createTimestamp(boolean doOffsetWorkaround) {
		Calendar shipDate = Calendar.getInstance();

		// This is a hack to workaround a problem with the FedEx API.
		// Regardless of the TimeZone actually in use by the Calendar object
		// the XML output uses GMT for the ship timestamp. So we subtract the
		// GMT offset from the Calendar object first so it outputs correctly.
		Calendar gmtDate = Calendar.getInstance();
		if (doOffsetWorkaround) {
			gmtDate.setTimeZone(TimeZone.getTimeZone("GMT"));
			int gmtHour = gmtDate.get(Calendar.HOUR_OF_DAY);
			int localHour = shipDate.get(Calendar.HOUR_OF_DAY);
			if (gmtHour < localHour) {
				gmtHour = gmtHour + 24;
			}
			int offset = gmtHour - localHour;
			shipDate.add(Calendar.HOUR_OF_DAY, -offset);
			if (log.isInfoEnabled()) {
				log.info("Adjusted ship timestamp hour (FedEx API workaround): " + shipDate.get(Calendar.HOUR_OF_DAY));
			}
		}

		// Add extra hours to the ship timestamp
		String fedExExtraDeliveryHours = (String) getSettings().getValue("fedExExtraDeliveryHours");
		if (fedExExtraDeliveryHours != null && fedExExtraDeliveryHours.length() > 0) {
			try {
				int extraHours = Integer.parseInt(fedExExtraDeliveryHours);
				shipDate.add(Calendar.HOUR_OF_DAY, extraHours);
				if (log.isInfoEnabled())
					log.info("Added: " + extraHours + " extra hours to ship timestamp (now "
							+ shipDate.get(Calendar.HOUR_OF_DAY) + ")");
			} catch (NumberFormatException nfe) {
				if (log.isWarnEnabled())
					log.warn("Invalid value for fedExExtraDeliveryHours: " + fedExExtraDeliveryHours);
			}
		}

		// If we're starting on a weekend or holiday, push past those
		shipDate = pushPastWeekendsAndHolidays(shipDate);

		// Add extra days to the ship timestamp
		String extraDays = (String) getSettings().getValue("fedExExtraDeliveryDays");
		if (extraDays != null && extraDays.length() > 0) {
			for (int i = 0; i < Integer.parseInt(extraDays); i++) {
				shipDate.add(Calendar.DATE, 1);
				log.info("Adding fedExExtraDeliveryDays number " + (i + 1) + " of " + extraDays
						+ ", delivery date is now: " + formatDeliveryDate(shipDate));
				shipDate = pushPastWeekendsAndHolidays(shipDate);
			}
		}
		return shipDate;
	}

	protected DropoffType createDropoffType() {
		String fedExDropoffType = (String) getSettings().getValue("fedExDropoffType");
		if (fedExDropoffType == null || fedExDropoffType.equals("REGULAR_PICKUP")) {
			return DropoffType.REGULAR_PICKUP;
		} else if (fedExDropoffType.equals("STATION")) {
			return DropoffType.STATION;
		} else if (fedExDropoffType.equals("REQUEST_COURIER")) {
			return DropoffType.REQUEST_COURIER;
		} else if (fedExDropoffType.equals("DROP_BOX")) {
			return DropoffType.DROP_BOX;
		} else if (fedExDropoffType.equals("BUSINESS_SERVICE_CENTER")) {
			return DropoffType.BUSINESS_SERVICE_CENTER;
		} else {
			return DropoffType.REGULAR_PICKUP;
		}
	}

	protected PackagingType createPackagingType() {
		String fedExPackaging = (String) getSettings().getValue("fedExPackaging");
		if (fedExPackaging == null || fedExPackaging.equals("YOUR_PACKAGING")) {
			return PackagingType.YOUR_PACKAGING;
		} else if (fedExPackaging.equals("FEDEX_10KG_BOX")) {
			return PackagingType.FEDEX_10KG_BOX;
		} else if (fedExPackaging.equals("FEDEX_25KG_BOX")) {
			return PackagingType.FEDEX_25KG_BOX;
		} else if (fedExPackaging.equals("FEDEX_BOX")) {
			return PackagingType.FEDEX_BOX;
		} else if (fedExPackaging.equals("FEDEX_ENVELOPE")) {
			return PackagingType.FEDEX_ENVELOPE;
		} else if (fedExPackaging.equals("FEDEX_PAK")) {
			return PackagingType.FEDEX_PAK;
		} else if (fedExPackaging.equals("FEDEX_TUBE")) {
			return PackagingType.FEDEX_TUBE;
		} else {
			return PackagingType.YOUR_PACKAGING;
		}
	}

	protected Party createShipperParty(Map parameters) {
		Party merchantParty = new Party();
		Contact merchantContact = new Contact();
		Address merchantAddress = new Address();
		merchantAddress.setStateOrProvinceCode((String) parameters.get("fedExOriginAddressStateOrProvinceCode"));
		merchantAddress.setPostalCode((String) parameters.get("fedExOriginAddressPostalCode"));
		merchantAddress.setCountryCode((String) parameters.get("fedExOriginAddressCountryCode"));
		merchantParty.setContact(merchantContact);
		merchantParty.setAddress(merchantAddress);
		return merchantParty;
	}

	protected Party createRecipientParty(Map parameters) {
		Party customerParty = new Party();
		Contact customerContact = new Contact();
		Address customerAddress = new Address();
		customerAddress.setCity(getUser().getOrder().firstDelivery().getCity());
		customerAddress.setStateOrProvinceCode(getUser().getOrder().firstDelivery().getState());
		customerAddress.setPostalCode(getUser().getOrder().firstDelivery().getPostalCode());
		customerAddress.setCountryCode(getUser().getOrder().firstDelivery().getCountry());

		// Residential flag
		String fedExResidence = (String) getSettings().getValue("fedExResidence");
		boolean isResidential = false;
		if (fedExResidence != null && fedExResidence.equals("RESIDENTIAL")) {
			isResidential = true;
		}
		if (fedExResidence != null
				&& fedExResidence.equals("ORGANIZATION_FIELD")
				&& (getUser().getOrder().firstDelivery().getOrganization() == null || getUser().getOrder()
						.firstDelivery().getOrganization().trim().length() == 0)) {
			isResidential = true;
		}
		if (getUser().getOrder() != null && getUser().getOrder().firstDelivery() != null
				&& getUser().getOrder().firstDelivery().getIsResidential() != null) {
			isResidential = getUser().getOrder().firstDelivery().getIsResidential().booleanValue();
		}
		if (parameters != null && parameters.get("fedExResidence") != null) {
			isResidential = ((Boolean) parameters.get("fedExResidence")).booleanValue();
		}
		customerAddress.setResidential(Boolean.valueOf(isResidential));
		customerParty.setContact(customerContact);
		customerParty.setAddress(customerAddress);
		return customerParty;
	}

	protected Payment createShippingChargesPayment() {
		Payment shippingChargesPayment = new Payment();
		shippingChargesPayment.setPaymentType(PaymentType.SENDER);
		return shippingChargesPayment;
	}

	protected PackageSpecialServicesRequested createSignatureOption() {
		SignatureOptionType sot = SignatureOptionType.SERVICE_DEFAULT;
		String fedExDefaultSignatureOption = (String) getSettings().getValue("fedExDefaultSignatureOption");
		if (fedExDefaultSignatureOption != null) {
			fedExDefaultSignatureOption = fedExDefaultSignatureOption.trim();
		}
		if (fedExDefaultSignatureOption == null || fedExDefaultSignatureOption.equals("")
				|| fedExDefaultSignatureOption.equals("SERVICE_DEFAULT")) {
			sot = SignatureOptionType.SERVICE_DEFAULT;
		} else if (fedExDefaultSignatureOption.equals("ADULT")) {
			sot = SignatureOptionType.ADULT;
		} else if (fedExDefaultSignatureOption.equals("INDIRECT")) {
			sot = SignatureOptionType.INDIRECT;
		} else if (fedExDefaultSignatureOption.equals("DIRECT")) {
			sot = SignatureOptionType.DIRECT;
		} else if (fedExDefaultSignatureOption.equals("NO_SIGNATURE_REQUIRED")) {
			sot = SignatureOptionType.NO_SIGNATURE_REQUIRED;
		}
		SignatureOptionDetail sod = new SignatureOptionDetail();
		sod.setOptionType(sot);
		PackageSpecialServicesRequested pssr = new PackageSpecialServicesRequested();
		pssr.setSignatureOptionDetail(sod);
		PackageSpecialServiceType psst = PackageSpecialServiceType.SIGNATURE_OPTION;
		pssr.setSpecialServiceTypes(new PackageSpecialServiceType[] { psst });
		return pssr;
	}

	protected Collection createPackages(Map parameters) {
		Collection packages = new ArrayList();
		if (getUser().getOrder().getOrderDeliveryCollection() != null
				&& !getUser().getOrder().getOrderDeliveryCollection().isEmpty()) {
			int count = 0;
			Iterator j = getUser().getOrder().getOrderShippingRules().iterator();
			while (j.hasNext()) {
				OrderShippingRule osr = (OrderShippingRule) j.next();
				boolean createThePackage = isPackageOkToCreate(osr);
				if (createThePackage) {
					if (log.isInfoEnabled())
						log.info("Processing package: " + osr.getCode());
					RequestedPackageLineItem requestedPackage = createPackage(parameters, osr, count++);
					if (requestedPackage != null) {
						packages.add(requestedPackage);
					}
				}
			}
		}
		return packages;
	}

	public RequestedPackageLineItem createPackage(Map parameters,
			OrderShippingRule osr, int count) {
		RequestedPackageLineItem rpli = new RequestedPackageLineItem();

		// Call the superclass to get the package weight
		Double weightDouble = findPackageWeight(osr);

		// Package might have no items, so just return (skips creating the
		// package)
		if (weightDouble == null) {
			return null;
		}

		// Add the extra weight
		String fedExExtraWeight = (String) getSettings().getValue(
				"fedExExtraWeight");
		if (fedExExtraWeight != null) {
			weightDouble = new Double(weightDouble.doubleValue()
					+ Double.parseDouble(fedExExtraWeight));
		}

		// If weight is still zero, let's skip the package
		if (weightDouble.doubleValue() == 0) {
			return null;
		}
		Weight rpWeight = new Weight();
		rpWeight.setValue(new BigDecimal(weightDouble.doubleValue()));

		// Set the weight units
		String fedExWeightUnits = (String) getSettings().getValue(
				"fedExWeightUnits");
		if (fedExWeightUnits == null || fedExWeightUnits.equals("LB")) {
			rpWeight.setUnits(WeightUnits.LB);
		} else {
			rpWeight.setUnits(WeightUnits.KG);
		}
		rpli.setWeight(rpWeight);

		// Dimensions
		Integer length = null;
		Integer width = null;
		Integer height = null;
		String fedExDefaultLength = (String) getSettings().getValue(
				"fedExDefaultLength");
		String fedExDefaultWidth = (String) getSettings().getValue(
				"fedExDefaultWidth");
		String fedExDefaultHeight = (String) getSettings().getValue(
				"fedExDefaultHeight");

		// Use the package dimensions if they are defined, otherwise, use the
		// defaults if they are defined, otherwise, don't send dimensions.
		if (osr.getPkgLength() != null && osr.getPkgLength().doubleValue() > 0
				&& osr.getPkgWidth() != null
				&& osr.getPkgWidth().doubleValue() > 0
				&& osr.getPkgHeight() != null
				&& osr.getPkgHeight().doubleValue() > 0) {
			length = new Integer(osr.getPkgLength().intValue());
			width = new Integer(osr.getPkgWidth().intValue());
			height = new Integer(osr.getPkgHeight().intValue());
		} else if (fedExDefaultLength != null
				&& Integer.parseInt(fedExDefaultLength) > 0
				&& fedExDefaultWidth != null
				&& Integer.parseInt(fedExDefaultWidth) > 0
				&& fedExDefaultHeight != null
				&& Integer.parseInt(fedExDefaultHeight) > 0) {
			length = new Integer(fedExDefaultLength);
			width = new Integer(fedExDefaultWidth);
			height = new Integer(fedExDefaultHeight);
		}
		if (length != null && width != null && height != null && length > 0 && width > 0 && height > 0) {
			Dimensions rpDimensions = new Dimensions();
			rpDimensions.setLength(new NonNegativeInteger(length.toString()));
			rpDimensions.setWidth(new NonNegativeInteger(width.toString()));
			rpDimensions.setHeight(new NonNegativeInteger(height.toString()));

			// Set the dimension units
			String fedExDimensionsUnits = (String) getSettings().getValue(
					"fedExDimensionsUnits");
			if (fedExDimensionsUnits == null
					|| fedExDimensionsUnits.equals("IN")) {
				rpDimensions.setUnits(LinearUnits.IN);
			} else {
				rpDimensions.setUnits(LinearUnits.CM);
			}
			rpli.setDimensions(rpDimensions);
		}
		return rpli;
	}

	public SmartPostShipmentDetail createSmartPostShipmentDetail(String fedexSmartPostIndiciaType,
			String fedexSmartPostHubId) {
		SmartPostShipmentDetail spd = new SmartPostShipmentDetail();
		SmartPostIndiciaType spit = SmartPostIndiciaType.fromValue(fedexSmartPostIndiciaType);
		spd.setIndicia(spit);
		spd.setHubId(fedexSmartPostHubId);
		return spd;
	}

	private Map processRateResponse(Map parameters, RateReply reply)
			throws Exception {
		Map ratesAndErrors = new HashMap();
		Collection rates = new ArrayList();
		Map option = new HashMap();
		String storeCurrencyCode = (String) getSettings().getValue(
				"currencyCode");
		String fedExShowListPrices = (String) getSettings().getValue(
				"fedExShowListPrices");
		String fedExService = (String) getSettings().getValue("fedExService");
		RateReplyDetail[] rrds = reply.getRateReplyDetails();
		for (int i = 0; i < rrds.length; i++) {
			RateReplyDetail rrd = rrds[i];
			String serviceType = rrd.getServiceType().toString();
			if (!isInCommaSeparatedString(serviceType, fedExService)) {
				log.info("Skipping rate because it's not set to be shown: "
						+ serviceType);
				continue;
			}
			option = new HashMap();
			option.putAll(PropertyUtils.describe(rrd));
			option.put("code", rrd.getServiceType().toString());
			option.put("name", lookUpNameFromCode(rrd.getServiceType()
					.toString()));
			if (rrd.getDeliveryTimestamp() != null) {
				Date deliveryDate = new Date(rrd.getDeliveryTimestamp()
						.getTimeInMillis());
				String fullDate = formatDateTime(deliveryDate);
				option.put("fullDate", fullDate);
				option.put("DeliveryDate", fullDate.substring(0, 10));
			} else if (rrd.getTransitTime() != null) {
				String delivDate = deliveryDateFromTransitTimeType(rrd.getTransitTime());
				option.put("DeliveryDate", delivDate);
				if (rrd.getMaximumTransitTime() != null) {
					String maxDelivDate = deliveryDateFromTransitTimeType(rrd.getMaximumTransitTime());
					option.put("MaxDeliveryDate", maxDelivDate);
				} else if (rrd.getCommitDetails() != null && rrd.getCommitDetails().length == 1
						&& rrd.getCommitDetails(0).getMaximumTransitTime() != null) {
					String maxDelivDate = deliveryDateFromTransitTimeType(rrd.getCommitDetails(0)
							.getMaximumTransitTime());
					option.put("MaxDeliveryDate", maxDelivDate);
				}
			}
			RatedShipmentDetail[] rsds = rrd.getRatedShipmentDetails();
			for (int j = 0; j < rsds.length; j++) {
				RatedShipmentDetail rsd = rsds[j];
				ShipmentRateDetail srd = rsd.getShipmentRateDetail();
				if (fedExShowListPrices.equals("1")
						&& srd.getRateType().getValue().indexOf("_ACCOUNT_") != -1) {
					log.info("Skipping rate because fedExShowListPrices is"
							+ " set to yes and this rate is an account rate. Account rate for " + rrd.getServiceType()
							+ " is: " + srd.getTotalNetCharge().getAmount());
					continue;
				}
				if (fedExShowListPrices.equals("0")
						&& srd.getRateType().getValue().indexOf("_LIST_") != -1) {
					log.info("Skipping rate because fedExShowListPrices "
							+ "is set to no and this rate is a list rate. List rate for " + rrd.getServiceType()
							+ " is: " + srd.getTotalNetCharge().getAmount());
					continue;
				}
				String currencyCode = srd.getTotalNetCharge().getCurrency();
				if (!storeCurrencyCode.equals(currencyCode)) {
					log.warn("Skipping rate because its currency code ("
							+ currencyCode + ") does not match the store's ("
							+ storeCurrencyCode + ")");
					continue;
				}
				option.putAll(PropertyUtils.describe(srd));
				double totalChargesDouble = srd.getTotalNetCharge().getAmount()
						.doubleValue();
				option.put("priceDouble", new Double(totalChargesDouble));
				option.put("originalPriceDouble",
						new Double(totalChargesDouble));
				option.put("price", formatPrice(totalChargesDouble,
						currencyCode, getUser().getLocale()));
				option.put("originalPrice", formatPrice(totalChargesDouble,
						currencyCode, getUser().getLocale()));
				option.put("currencyCode", currencyCode);
			}
			option.put("processorClass", "FedExShippingProcessor");
			option.put("processorName", "FedEx");
			Collection discounts = new ArrayList();
			option.put("discounts", discounts);
			rates.add(option);
			if (log.isInfoEnabled()) {
				log.info("FedEx option: " + option);
			}
		}
		ratesAndErrors.put("rates", rates);
		return ratesAndErrors;
	}

	private boolean isResponseOk(
			NotificationSeverityType notificationSeverityType) {
		if (notificationSeverityType == null) {
			return false;
		}
		if (notificationSeverityType.equals(NotificationSeverityType.NOTE)
				|| notificationSeverityType
						.equals(NotificationSeverityType.SUCCESS)) {
			return true;
		}
		return false;
	}

	protected boolean isResponseBenignError(RateReply reply) {
		boolean ret = false;
		Notification[] notifications = reply.getNotifications();
		if (notifications != null && notifications.length > 0) {
			for (int i = 0; i < notifications.length; i++) {
				Notification n = notifications[i];
				NotificationSeverityType nst = n.getSeverity();
				if (!isResponseOk(nst) && n.getCode() != null && n.getCode().equals("608")) {
					log.warn("Benign error (ignoring); " + n.getCode() + ": " + n.getMessage());
					ret = true;
				} else if (!isResponseOk(nst)) {
					ret = false;
					break;
				}
			}
		}
		return ret;
	}

	private ClientDetail createClientDetail() {
		ClientDetail clientDetail = new ClientDetail();
		String accountNumber = (String) getSettings().getValue(
				"fedExAccountNumber");
		String meterNumber = (String) getSettings()
				.getValue("fedExMeterNumber");
		clientDetail.setAccountNumber(accountNumber);
		clientDetail.setMeterNumber(meterNumber);
		return clientDetail;
	}

	private WebAuthenticationDetail createWebAuthenticationDetail() {
		WebAuthenticationCredential wac = new WebAuthenticationCredential();
		String key = (String) getSettings().getValue("fedExAPIKey");
		String password = (String) getSettings().getValue("fedExPassword");
		wac.setKey(key);
		wac.setPassword(password);
		return new WebAuthenticationDetail(wac);
	}

	private Map processErrors(Notification[] notifications) {
		Map results = new HashMap();
		if (notifications == null || notifications.length == 0) {
			log.info("No notifications returned");
		}
		for (int i = 0; i < notifications.length; i++) {
			Notification n = notifications[i];
			NotificationSeverityType nst = n.getSeverity();
			StringBuffer sb = new StringBuffer("");
			sb.append(n.getMessage());
			sb.append(" (Severity: " + (nst == null ? "null" : nst.getValue()));
			sb.append(" Code: " + n.getCode());
			sb.append(" Source: " + n.getSource() + ")");
			if (nst.getValue() != null
					&& nst.equals(NotificationSeverityType.ERROR)) {
				log.warn(sb.toString());
				results.put("errorCode", n.getCode());
				results.put("errorMessage", n.getMessage());
			} else if (nst.getValue() != null
					&& nst.equals(NotificationSeverityType.FAILURE)) {
				log.error(sb.toString());
				results.put("errorCode", n.getCode());
				results.put("errorMessage", n.getMessage());
			} else if (nst.getValue() != null
					&& nst.equals(NotificationSeverityType.WARNING)) {
				results.put("errorCode", n.getCode());
				results.put("errorMessage", n.getMessage());
				log.warn(sb.toString());
			}
		}
		return results;
	}

	private void updateEndPoint(RateServiceLocator serviceLocator) {
		String endPoint = (String) getSettings().getValue("fedExURL");
		if (endPoint != null) {
			serviceLocator.setRateServicePortEndpointAddress(endPoint);
		}
	}

	public boolean isInCommaSeparatedString(String value,
			String commaSeparatedString) {
		boolean result = false;
		String[] separatedValues = commaSeparatedString.split(",");
		for (int j = 0; j < separatedValues.length; j++) {
			String thisValue = separatedValues[j].trim();
			if (value.equals(thisValue)) {
				result = true;
				break;
			}
		}
		return result;
	}

	public Collection doGlobalExtraCharge(Collection mergedRates) {
		String fedExExtraCharge = (String) getSettings().getValue(
				"fedExExtraCharge");
		if (mergedRates == null || mergedRates.isEmpty()
				|| fedExExtraCharge == null
				|| fedExExtraCharge.replaceAll("\\D", "").length() == 0) {
			return mergedRates;
		}
		String currencyCode = (String) getSettings().getValue("currencyCode");
		Iterator i = mergedRates.iterator();
		while (i.hasNext()) {
			Map mergedRate = (Map) i.next();
			Double mergedPriceDouble = (Double) mergedRate.get("priceDouble");
			if (mergedPriceDouble == null) {
				mergedPriceDouble = new Double(0);
			}
			double newPrice = mergedPriceDouble.doubleValue();
			try {
				String extraCharge = fedExExtraCharge;
				extraCharge = extraCharge.trim();
				if (extraCharge.endsWith("%")) {

					// Charge is a percentage
					extraCharge.replaceAll("\\D", "");
					double percentageCharge = Double.parseDouble(extraCharge);
					double doubleCharge = utils().roundForMoney(
							percentageCharge * newPrice).doubleValue();
					doubleCharge = doubleCharge / 100;
					newPrice = newPrice + doubleCharge;
				} else {

					// Charge is an amount
					extraCharge.replaceAll("\\D", "");
					newPrice = newPrice + Double.parseDouble(extraCharge);
				}
			} catch (NumberFormatException nfe) {
				if (log.isErrorEnabled())
					log.error("FedEx extra charge format error: "
							+ fedExExtraCharge, nfe);
			}
			mergedRate.put("priceDouble", new Double(newPrice));
			mergedRate.put("originalPriceDouble", new Double(newPrice));
			mergedRate.put("price", formatPrice(newPrice, currencyCode,
					getUser().getLocale()));
			mergedRate.put("originalPrice", formatPrice(newPrice, currencyCode,
					getUser().getLocale()));
		}
		return mergedRates;
	}

	public String missingRequiredSettings() {
		Collection requiredSettings = new ArrayList();
		requiredSettings.add("fedExURL");
		requiredSettings.add("fedExAccountNumber");
		requiredSettings.add("fedExMeterNumber");
		requiredSettings.add("fedExAPIKey");
		requiredSettings.add("fedExPassword");
		requiredSettings.add("fedExService");
		requiredSettings.add("fedExOriginAddressPostalCode");
		requiredSettings.add("fedExOriginAddressCountryCode");
		StringBuffer missingSettings = new StringBuffer("");
		Iterator i = requiredSettings.iterator();
		while (i.hasNext()) {
			String setting = (String) i.next();
			String value = (String) getSettings().getValue(setting);
			if (value == null || value.length() == 0) {
				if (missingSettings.length() > 0) {
					missingSettings.append(", ");
				}
				missingSettings.append(setting);
			}
		}
		return missingSettings.toString();
	}

	public String lookUpNameFromCode(String code) {
		String fedExServiceOptions = (String) getSettings().getValue(
				"fedExServiceOptions");
		String fedExServiceOptionNames = (String) getSettings().getValue(
				"fedExServiceOptionNames");
		String[] optionsArray = fedExServiceOptions.split(",");
		String[] namesArray = fedExServiceOptionNames.split(",");
		for (int j = 0; j < optionsArray.length; j++) {
			String thisValue = optionsArray[j].trim();
			if (thisValue.equals(code) && namesArray.length > j) {
				return namesArray[j].trim();
			}
		}
		return code;
	}

	protected String deliveryDateFromTransitTimeType(TransitTimeType transitTimeType) {
		String deliveryDate = null;
		int transitTimeInt = transitTimeIntFromTransitTimeType(transitTimeType);
		if (transitTimeInt > 0) {
			deliveryDate = doEstimatedDeliveryDate(createTimestamp(false), transitTimeInt);
		}
		return deliveryDate;
	}

	protected int transitTimeIntFromTransitTimeType(TransitTimeType transitTimeType) {
		int transitTimeInt = 0;
		if (transitTimeType != null && transitTimeType.getValue() != null) {
			if (transitTimeType.getValue().equals(TransitTimeType.ONE_DAY.getValue())) {
				transitTimeInt = 1;
			} else if (transitTimeType.getValue().equals(TransitTimeType.TWO_DAYS.getValue())) {
				transitTimeInt = 2;
			} else if (transitTimeType.getValue().equals(TransitTimeType.THREE_DAYS.getValue())) {
				transitTimeInt = 3;
			} else if (transitTimeType.getValue().equals(TransitTimeType.FOUR_DAYS.getValue())) {
				transitTimeInt = 4;
			} else if (transitTimeType.getValue().equals(TransitTimeType.FIVE_DAYS.getValue())) {
				transitTimeInt = 5;
			} else if (transitTimeType.getValue().equals(TransitTimeType.SIX_DAYS.getValue())) {
				transitTimeInt = 6;
			} else if (transitTimeType.getValue().equals(TransitTimeType.SEVEN_DAYS.getValue())) {
				transitTimeInt = 7;
			} else if (transitTimeType.getValue().equals(TransitTimeType.EIGHT_DAYS.getValue())) {
				transitTimeInt = 8;
			} else if (transitTimeType.getValue().equals(TransitTimeType.NINE_DAYS.getValue())) {
				transitTimeInt = 9;
			} else if (transitTimeType.getValue().equals(TransitTimeType.TEN_DAYS.getValue())) {
				transitTimeInt = 10;
			} else if (transitTimeType.getValue().equals(TransitTimeType.ELEVEN_DAYS.getValue())) {
				transitTimeInt = 11;
			} else if (transitTimeType.getValue().equals(TransitTimeType.TWELVE_DAYS.getValue())) {
				transitTimeInt = 12;
			} else if (transitTimeType.getValue().equals(TransitTimeType.THIRTEEN_DAYS.getValue())) {
				transitTimeInt = 13;
			} else if (transitTimeType.getValue().equals(TransitTimeType.FOURTEEN_DAYS.getValue())) {
				transitTimeInt = 14;
			} else if (transitTimeType.getValue().equals(TransitTimeType.FIFTEEN_DAYS.getValue())) {
				transitTimeInt = 15;
			} else if (transitTimeType.getValue().equals(TransitTimeType.SIXTEEN_DAYS.getValue())) {
				transitTimeInt = 16;
			} else if (transitTimeType.getValue().equals(TransitTimeType.SEVENTEEN_DAYS.getValue())) {
				transitTimeInt = 17;
			} else if (transitTimeType.getValue().equals(TransitTimeType.EIGHTEEN_DAYS.getValue())) {
				transitTimeInt = 18;
			} else if (transitTimeType.getValue().equals(TransitTimeType.NINETEEN_DAYS.getValue())) {
				transitTimeInt = 19;
			} else if (transitTimeType.getValue().equals(TransitTimeType.TWENTY_DAYS.getValue())) {
				transitTimeInt = 20;
			}
		}
		return transitTimeInt;
	}
}
