package com.softslate.commerce.customer.core;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.softslate.commerce.businessobjects.core.Settings;

/**
 * Filter to translate incoming URLs from a search-engine friendly format to the
 * format understood by the application. Various settings in the application
 * tell the system how to do the translating. For example, if the
 * <code>seoProductMarker</code> setting is set to "product", it tells us to
 * look for the pattern "/product/&lt;product-code&gt;" in the search-engine
 * friendly URL, which would be translated to
 * "/Product.do?code=&lt;product-code&gt;";
 * <p>
 * If the incoming URL is determined to be in a search-engine friendly format
 * and the translation is successful, the request is forwarded to the translated
 * URL. Otherwise, it is simply passed on to the next filter (if any).
 * <p>
 * URL paths ending in ".do", as well as common image, stylesheet, and
 * javascript file extensions, are immediately passed through to the next
 * filter.
 * 
 * @author David Tobey
 * 
 */
public class SEOFilter extends BaseFilter implements Filter {
	static Log log = LogFactory.getLog(SEOFilter.class);

	public static final String welcomePath = "/Welcome.do";

	public static final String productPath = "/Product.do";

	public static final String productListPath = "/ProductList.do";

	public static final String categoryPath = "/Category.do";

	public static final String manufacturerPath = "/Manufacturer.do";

	public static final String manufacturersPath = "/Manufacturers.do";

	public static final String pageEndPath = "Page.do";

	public static final String contentPath = "/Content.do";

	private FilterConfig filterConfig = null;

	public void init(FilterConfig filterConfig) throws ServletException {
		this.filterConfig = filterConfig;
	}

	public void destroy() {
	}

	public void doFilter(ServletRequest request, ServletResponse response,
			FilterChain chain) throws IOException, ServletException {

		// Must cast incoming request to HttpServletRequest
		HttpServletRequest req = (HttpServletRequest) request;

		// Pass the request along right away if it ends in .do, or is an image.
		if (req.getServletPath().startsWith("/administrator/") || req.getServletPath().startsWith("/api/")
				|| req.getServletPath().endsWith(".do") || req.getServletPath().endsWith(".jsp")
				|| isStaticRequest(req)) {
			chain.doFilter(request, response);
			return;
		}

		// Try grabbing the settings from the application scope; forward to
		// Welcome.do if we can't.
		Settings settings = (Settings) filterConfig.getServletContext()
				.getAttribute("settings");
		if (settings == null) {
			if (log.isInfoEnabled()) {
				log
						.info("Can't perform SEO translation for "
								+ req.getServletPath()
								+ req.getQueryString()
								+ " because the application's settings have not been initialized.");
				log.info("Forwarding to welcome screen.");
			}
			RequestDispatcher rd = request.getRequestDispatcher(welcomePath);
			rd.forward(request, response);
			return;
		}
		// Chances are high we're dealing with a SEO-friendly URL, so let's
		// translate.
		if (log.isDebugEnabled())
			log.debug("-------Translating path: " + req.getServletPath()
					+ "--------");
		String forwardPath = translateSEOURL(settings, req.getServletPath());
		if (forwardPath != null) {
			// Append the query string, if there was one.
			if (req.getQueryString() != null) {
				if (forwardPath.indexOf("?") != -1) {
					forwardPath = forwardPath + "&" + req.getQueryString();
				} else {
					forwardPath = forwardPath + "?" + req.getQueryString();
				}
			}
			// Save the HTTP refer so we can log it in BaseRequestProcessor
			if (req.getHeader("Referer") != null) {
				req.setAttribute("Referer", req.getHeader("Referer"));
			}
			// Forward the request.
			if (log.isInfoEnabled())
				log.info(req.getServletPath() + " translated to: "
						+ forwardPath);
			RequestDispatcher rd = request.getRequestDispatcher(forwardPath);
			rd.forward(request, response);
			return;
		}
		// If we can't figure out the URL, just pass it along.
		if (log.isInfoEnabled())
			log.info("Failed to translate path: " + req.getServletPath());
		chain.doFilter(request, response);
		return;
	}

	/**
	 * Translates a given URL path from a search-engine friendly format to a
	 * format recognized by the application. Uses the given settings to
	 * determine how to translate. Translations are done for the Welcome screen,
	 * Page screens, Product Screen, and Category Screen.
	 * 
	 * @param settings
	 *            A Bean holding the settings that determine how to translate.
	 * @param servletPath
	 *            The incoming URL path (from after the context path until the
	 *            query string)
	 * @return The translated URL.
	 */
	public String translateSEOURL(Settings settings, String servletPath) {
		String seoWelcomeMarker = (String) settings
				.getValue("seoWelcomeMarker");
		String seoPageMarker = (String) settings.getValue("seoPageMarker");
		String seoContentMarker = (String) settings.getValue("seoContentMarker");
		String seoProductMarker = (String) settings
				.getValue("seoProductMarker");
		String seoCategoryMarker = (String) settings
				.getValue("seoCategoryMarker");
		String seoManufacturerMarker = (String) settings
				.getValue("seoManufacturerMarker");
		String seoProductListMarker = (String) settings
		.getValue("seoProductListMarker");
		String seoSuffix = (String) settings.getValue("seoSuffix");
		/*
		 * if (log.isDebugEnabled()) { log.debug("seoWelcomeMarker: " +
		 * seoWelcomeMarker); log.debug("seoPageMarker: " + seoPageMarker);
		 * log.debug("seoProductMarker: " + seoProductMarker);
		 * log.debug("seoCategoryMarker: " + seoCategoryMarker);
		 * log.debug("seoSuffix: " + seoSuffix); }
		 */
		// First check for the welcome URL
		if (seoWelcomeMarker.length() > 0
				&& servletPath.indexOf("/" + seoWelcomeMarker) != -1) {
			return welcomePath;
		}
		// Check for page URLs
		if (seoPageMarker.length() > 0
				&& servletPath.indexOf("/" + seoPageMarker + "/") != -1) {
			String pageOrCode = extractPageOrCode(servletPath, seoPageMarker,
					seoSuffix);
			return "/" + pageOrCode + pageEndPath;
		}
		// Check for content URLs
		if (seoContentMarker != null && seoContentMarker.length() > 0
				&& servletPath.indexOf("/" + seoContentMarker + "/") != -1) {
			String firstRow = "";
			String otherParams = "";
			String categoryCode = extractPageOrCode(servletPath,
					seoContentMarker, seoSuffix);
			int anotherSlash = categoryCode.indexOf("/");
			if (anotherSlash != -1) {
				firstRow = categoryCode.substring(anotherSlash + 1);
				categoryCode = categoryCode.substring(0, anotherSlash);
			}
			int otherParamsStart = firstRow.indexOf("?");
			if (otherParamsStart != -1) {
				otherParams = firstRow.substring(otherParamsStart);
				firstRow = firstRow.substring(0, otherParamsStart);
			}
			String url = contentPath + "?code=" + categoryCode;
			if (firstRow.length() > 0) {
				url = url + "&firstRow=" + firstRow;
			}
			if (otherParams.length() > 0) {
				url = url + otherParams;
			}
			return url;
		}
		// Check for category URLs
		if (seoCategoryMarker.length() > 0
				&& servletPath.indexOf("/" + seoCategoryMarker + "/") != -1) {
			String firstRow = "";
			String otherParams = "";
			String categoryCode = extractPageOrCode(servletPath,
					seoCategoryMarker, seoSuffix);
			int anotherSlash = categoryCode.indexOf("/");
			if (anotherSlash != -1) {
				firstRow = categoryCode.substring(anotherSlash + 1);
				categoryCode = categoryCode.substring(0, anotherSlash);
			}
			int otherParamsStart = firstRow.indexOf("?");
			if (otherParamsStart != -1) {
				otherParams = firstRow.substring(otherParamsStart);
				firstRow = firstRow.substring(0, otherParamsStart);
			}
			String url = categoryPath + "?code=" + categoryCode;
			if (firstRow.length() > 0) {
				url = url + "&firstRow=" + firstRow;
			}
			if (otherParams.length() > 0) {
				url = url + otherParams;
			}
			return url;
		}
		// Check for product URLs
		if (seoProductMarker != null && seoProductMarker.length() > 0
				&& servletPath.indexOf("/" + seoProductMarker + "/") != -1) {
			String categoryCode = "";
			String productCode = extractPageOrCode(servletPath,
					seoProductMarker, seoSuffix);
			// Product URLs can have a category code shoved in there
			int anotherSlash = productCode.indexOf("/");
			if (anotherSlash != -1) {
				categoryCode = productCode.substring(0, anotherSlash);
				productCode = productCode.substring(anotherSlash + 1);
			}
			String url = productPath + "?code=" + productCode;
			if (categoryCode.length() > 0) {
				url = url + "&category=" + categoryCode;
			}
			return url;
		}
		// Check for product list page
		if (seoProductListMarker != null && seoProductListMarker.length() > 0
				&& servletPath.indexOf("/" + seoProductListMarker) != -1) {
			String otherParams = "";
			String firstRow = extractPageOrCode(servletPath,
					seoProductListMarker, seoSuffix);
			int otherParamsStart = firstRow.indexOf("?");
			if (otherParamsStart != -1) {
				otherParams = firstRow.substring(otherParamsStart);
				firstRow = firstRow.substring(0, otherParamsStart);
			}
			String url = productListPath;
			if (firstRow.length() > 0) {
				url = url + "?firstRow=" + firstRow;
			}
			if (otherParams.length() > 0) {
				otherParams = otherParams.replaceFirst("\\?", "&");
				url = url + otherParams;
			}
			return url;
		}
		// Check for manufacturer URLs
		if (seoManufacturerMarker != null && seoManufacturerMarker.length() > 0
				&& servletPath.indexOf("/" + seoManufacturerMarker + "/") != -1) {
			String firstRow = "";
			String otherParams = "";
			String manufacturerCode = extractPageOrCode(servletPath,
					seoManufacturerMarker, seoSuffix);
			if (manufacturerCode.length() == 0) {
				String url = manufacturersPath;
				return url;
			}
			int anotherSlash = manufacturerCode.indexOf("/");
			if (anotherSlash != -1) {
				firstRow = manufacturerCode.substring(anotherSlash + 1);
				manufacturerCode = manufacturerCode.substring(0, anotherSlash);
			}
			int otherParamsStart = firstRow.indexOf("?");
			if (otherParamsStart != -1) {
				otherParams = firstRow.substring(otherParamsStart);
				firstRow = firstRow.substring(0, otherParamsStart);
			}
			String url = manufacturerPath + "?code=" + manufacturerCode;
			if (firstRow.length() > 0) {
				url = url + "&firstRow=" + firstRow;
			}
			if (otherParams.length() > 0) {
				url = url + otherParams;
			}
			return url;
		}
		// Check for manufacturer list page
		if (seoManufacturerMarker != null && seoManufacturerMarker.length() > 0
				&& servletPath.indexOf("/" + seoManufacturerMarker) != -1) {
			String url = manufacturersPath;
			return url;
		}
		// Can't figure out pattern; return null
		return null;
	}

	/**
	 * Extracts the page identifier or the product, category, or manufacturer
	 * code from a given search-engine friendly URL.
	 * 
	 * @param servletPath
	 *            The URL to be parsed.
	 * @param marker
	 *            The marker identifying where to start looking for the page or
	 *            code.
	 * @param suffix
	 *            The suffix identifying where to stop looking for the page or
	 *            code.
	 * @return A string representing the page or code.
	 */
	public String extractPageOrCode(String servletPath, String marker,
			String suffix) {
		String pageOrCode = "";
		int startMarker = servletPath.indexOf("/" + marker + "/");
		if (startMarker == -1) {
			return "";
		}
		int start = startMarker + marker.length() + 2;
		int startSuffix = servletPath.lastIndexOf(suffix);
		if (suffix.length() == 0 || startSuffix == -1) {
			pageOrCode = servletPath.substring(start);
		} else {
			pageOrCode = servletPath.substring(start, startSuffix);
		}
		if (log.isDebugEnabled())
			log.debug("pageOrCode: " + pageOrCode);
		return pageOrCode;
	}
}
