package com.softslate.commerce.customer.customer;

import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import com.softslate.commerce.businessobjects.customer.Customer;
import com.softslate.commerce.businessobjects.customer.CustomerProcessor;
import com.softslate.commerce.businessobjects.email.Email;
import com.softslate.commerce.businessobjects.email.EmailSender;
import com.softslate.commerce.customer.core.ActionUtils;
import com.softslate.commerce.customer.core.BaseAction;
import com.softslate.commerce.customer.core.BaseForm;

/**
 * Struts action class used for requests to email a lost password to a customer. Used for requests to
 * "/LostPassword.do".
 * <p>
 * The <code>execute</code> method of this class invokes
 * {@link com.softslate.commerce.businessobjects.customer.CustomerProcessor#loadCustomerLostPassword(Map)} to look up
 * the customer information. If the customer is not found in the database or an email for the account is not found, the
 * <code>ActionForward</code> named "failure" will be returned.
 * <p>
 * If a password and an email address is found, the <code>execute</code> method will use
 * {@link com.softslate.commerce.customer.core.ActionUtils#sendMultipartEmail(String, String, String, String, String, String, String, String, String, String)}
 * to send the lost password email. A setting in the database named "lostPasswordEmailTemplate" is used as the template
 * for the text of the email. (See the Settings -&gt; Store screen of the administrator.)
 * 
 * @author David Tobey, jmcwill
 */
public class LostPasswordAction extends BaseAction {
	static Log log = LogFactory.getLog(LostPasswordAction.class);

	public ActionForward execute(ActionMapping mapping, ActionForm form,
			HttpServletRequest request, HttpServletResponse response)
			throws Exception {
		if (log.isInfoEnabled())
			log.info("Starting action.");
		BaseForm baseForm = (BaseForm) form;
		CustomerProcessor customerProcessor = (CustomerProcessor) baseForm.getBusinessObjectFactory().createObject(
				"customerProcessorImplementer");
		String useAdvancedLostPassword = (String) baseForm.getSettings().getValue("useAdvancedLostPassword");
		Map result = null;
		if ("1".equals(useAdvancedLostPassword)) {
			result = customerProcessor.loadCustomerAdvancedLostPassword(PropertyUtils.describe(baseForm));
		}else{
			result = customerProcessor.loadCustomerLostPassword(PropertyUtils.describe(baseForm));
		}
		if (result != null && result.get("resultCode") != null && !result.get("resultCode").equals("0")) {
			String resultCode = (String) result.get("resultCode");
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage(resultCode));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			if (log.isInfoEnabled() && !baseForm.getErrors().isEmpty())
				log.info("Validation error(s): " + baseForm.getErrors().toString());
			return mapping.findForward("failure");
		}

		Customer customer = (Customer) result.get("customer");
		if (customer == null || customer.getDecryptedPassword() == null) {
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.passwordNotFound"));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			return mapping.findForward("failure");
		}

		// Just in case something goes wrong with the token creation
		if ("1".equals(useAdvancedLostPassword) && StringUtils.isBlank(customer.getPasswordToken())) {
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.passwordNotFound"));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			return mapping.findForward("failure");
		}

		String validEmail = null;
		if (baseForm.formUtils().isValidEmail(customer.getUserName())) {
			validEmail = customer.getUserName();
		} else if (baseForm.formUtils().isValidEmail(customer.getEmail1())) {
			validEmail = customer.getEmail1();
		} else {
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.validEmailNotFound"));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			return mapping.findForward("failure");
		}

		try {
			if ("1".equals(useAdvancedLostPassword)
					&& StringUtils.isBlank(baseForm.getSettings().getValue("lostPasswordEmailSubjectAdvanced"))) {

				// Use a Template (newest implementation)
				customerProcessor.sendLostPasswordAdvancedEmail(customer);
			} else {
				ActionUtils actionUtils = (ActionUtils) baseForm.getBusinessObjectFactory()
						.createUtilsObject("actionUtilsImplementer");
				String subject = null;
				String textMessage = null;
				String htmlMessage = null;
				if ("1".equals(useAdvancedLostPassword)) {

					// If the settings are defined for the advanced method, use those (older implementation)
					subject = (String) baseForm.getSettings().getValue("lostPasswordEmailSubjectAdvanced");
					textMessage = (String) baseForm.getSettings().getValue("lostPasswordEmailTextTemplateAdvanced");
					htmlMessage = (String) baseForm.getSettings().getValue("lostPasswordEmailHTMLTemplateAdvanced");
					subject = actionUtils.parseLostPasswordTemplateAdvanced(customer, baseForm, subject,
							customer.getPasswordToken());
					textMessage = actionUtils.parseLostPasswordTemplateAdvanced(customer, baseForm, textMessage,
							customer.getPasswordToken());
					htmlMessage = actionUtils.parseLostPasswordTemplateAdvanced(customer, baseForm, htmlMessage,
							customer.getPasswordToken());
				} else {

					// Use legacy settings if the store is still set not to use the advanced method (oldest
					// implementation)
					subject = (String) baseForm.getSettings().getValue("lostPasswordEmailSubject");
					textMessage = (String) baseForm.getSettings().getValue("lostPasswordEmailTextTemplate");
					htmlMessage = (String) baseForm.getSettings().getValue("lostPasswordEmailHTMLTemplate");
					subject = actionUtils.parseLostPasswordTemplate(customer, baseForm, subject);
					textMessage = actionUtils.parseLostPasswordTemplate(customer, baseForm, textMessage);
					htmlMessage = actionUtils.parseLostPasswordTemplate(customer, baseForm, htmlMessage);
				}

				log.info("Sending email: " + subject);
				Email email = baseForm.getBusinessObjectFactory().createObject(Email.class);
				email.setCustomer(customer);
				email.setToEmails(validEmail);
				email.setType("lostPasswordEmail-" + customer.getCustomerID());
				email.setBodyHtml(htmlMessage);
				email.setBodyText(textMessage);
				email.setSubject(subject);
				EmailSender sender = baseForm.getBusinessObjectFactory().createObject(EmailSender.class);
				sender.setEmail(email);
				sender.send();
			}

		} catch (Exception e) {
			if (log.isErrorEnabled())
				log.error(e.toString(), e);
			baseForm.getErrors().add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.sendingEmail"));
			saveErrors(baseForm.getRequest(), (ActionMessages) baseForm.getErrors());
			return mapping.findForward("failure");
		}

		if ("1".equals(useAdvancedLostPassword)) {
			baseForm.getMessages().add(ActionMessages.GLOBAL_MESSAGE,
					new ActionMessage("messages.lostPasswordEmailedAdvanced"));
		} else {
			baseForm.getMessages()
					.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("messages.lostPasswordEmailed"));
		}

		saveMessages(baseForm.getRequest(), baseForm.getMessages());
		return mapping.findForward("success");

	}
}
